package com.sillycycle.bagleyd.skewb;

/*
 * @(#)Skewb.java
 *
 * Copyright 1994 - 2023  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>Skewb</code> class extends the
 * <code>SkewbCanvas</code> for handling the responses to the user
 * for the <code>SkewbFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

/*
The Simple Solutions to Cubic Puzzles by James G. Nourse, Bantam Books,
New York, 1981. pp 8-15.

Mastering the Magic Pyramid by Tom Werneck, Evans Brothers Limited, London,
1981. pp 109-111.

Beyond Rubik's Cube: spheres, pyramids, dodecahedrons and God knows what
else by Douglas R. Hofstadter, Scientific American, July 1982, pp 16-31.

Magic Cubes 1996 Catalog of Dr. Christoph Bandelow.
*/

public class Skewb extends Canvas implements MouseWheelListener,
		MouseListener, MouseMotionListener, KeyListener {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final int MAX_ITERATIONS = 100;

	static final int IGNORE = -1;
	static final int TR = 0;
	static final int BR = 1;
	static final int BL = 2;
	static final int TL = 3;
	static final int STRT = 4;
	static final int CW = 5;
	static final int HALF = 6;
	static final int CCW = 7;
	static final int TOP = 8;
	static final int RIGHT = 9;
	static final int BOTTOM = 10;
	static final int LEFT = 11;
	static final int MAX_FACES = 6;
	static final int MAX_ORIENT = 4;
	static final int MAX_ROTATE = 3;
	static final int MAX_CUBES = (MAX_ORIENT + 1);
	static final int MINOR = 0;
	static final int MAJOR = 1;

	static final int NORMAL = 1;
	static final int DOUBLE = 2;
	static final int INSTANT = 3;
	static final String DATAFILE = "skewb.dat";
	static final String SYMBOL = ":";

	Color       foreground, background;
	Color       borderColor;
	Color       currentForeground;
	Color[]     faceColor = new Color[MAX_FACES];
	int[]       directionOfPosition = null;
	int         currentDirection = -1;
	int         currentFace = -1, nextFace = -1;
	int         currentPosition = -1, nextPosition = -1;
	int         currentRotate = -1;
	int         currentControl = -1, nextControl = -1;
	int         rowPosition;
	boolean     started, cheat;
	boolean     mono, reverse, orientChoice, practiceChoice;
	int         delay;
	int         dim;
	int         lastX, lastY;

	JFrame      frame = null;
	boolean     mouseDown = false;
	Random      generator;
	transient SkewbStack  undo = null, redo = null;
	transient SkewbSolve  solve = null;
	Skewb2D     skewb2D = null;
	Skewb3D     skewb3D = null;
	String      stringSave = null;
	String[]    token = null;
	String      faceNames = null;

	static final int[][] facesToDirection/*[MAX_FACES][MAX_FACES]*/ = {
		{-1, BOTTOM, BOTTOM, BOTTOM, -1, TOP},
		{RIGHT, -1, RIGHT, -1, RIGHT, RIGHT},
		{TOP, LEFT, -1, RIGHT, BOTTOM, -1},
		{LEFT, -1, LEFT, -1, LEFT, LEFT},
		{-1, TOP, TOP, TOP, -1, BOTTOM},
		{BOTTOM, RIGHT, -1, LEFT, TOP, -1}
	};

	static final int[][][] facestoPosition/*[MAX_FACES][MAX_FACES][MAX_ORIENT]*/ = {
		{
			{-1, -1, -1, -1},
			{-1, -1, 0, 3},
			{-1, 0, 3, -1},
			{0, 3, -1, -1},
			{-1, -1, -1, -1},
			{1, -1, -1, 2}
		},
		{
			{2, -1, -1, 3},
			{-1, -1, -1, -1},
			{3, 2, -1, -1},
			{-1, -1, -1, -1},
			{-1, 3, 2, -1},
			{-1, -1, 3, 2}
		},
		{
			{1, -1, -1, 2},
			{-1, -1, 1, 0},
			{-1, -1, -1, -1},
			{3, 2, -1, -1},
			{-1, 0, 3, -1},
			{-1, -1, -1, -1}
		},
		{
			{0, -1, -1, 1},
			{-1, -1, -1, -1},
			{-1, -1, 1, 0},
			{-1, -1, -1, -1},
			{-1, 1, 0, -1},
			{1, 0, -1, -1}
		},
		{
			{-1, -1, -1, -1},
			{-1, -1, 2, 1},
			{1, -1, -1, 2},
			{2, 1, -1, -1},
			{-1, -1, -1, -1},
			{-1, 0, 3, -1}
		},
		{
			{-1, 0, 3, -1},
			{-1, -1, 3, 2},
			{-1, -1, -1, -1},
			{1, 0, -1, -1},
			{1, -1, -1, 2},
			{-1, -1, -1, -1}
		}
	};

	static final int[][] faceToRotate/*[MAX_FACES][MAX_ORIENT]*/ = {
		{3, 2, 1, 5},
		{2, 4, 5, 0},
		{3, 4, 1, 0},
		{5, 4, 2, 0},
		{3, 5, 1, 2},
		{3, 0, 1, 4}
	};

	transient SkewbLoc[][][] slideNextRow = new
		SkewbLoc[MAX_FACES][MAX_ORIENT][MAX_ORIENT / 2];
	transient SkewbLoc[][] minToMaj = new SkewbLoc[MAX_FACES][MAX_ORIENT];
	transient SkewbLoc[][] slideNextFace = new SkewbLoc[MAX_FACES][MAX_ORIENT];
	transient SkewbLocPos[][][] orthToDiag =
		new SkewbLocPos[MAX_FACES][MAX_ORIENT][MAX_ORIENT];
	transient SkewbLoc[] rotateToRow = new SkewbLoc[MAX_FACES];
	transient SkewbLoc[][] cubeLoc = null;
	static transient SkewbLoc[][] startLoc = null;
	transient SkewbLoc[] faceLoc = null;
	transient SkewbLoc[][] rowLoc = null;
	transient SkewbLoc[] minorLoc = null;
	transient SkewbLoc[][] majorLoc = null;

	public Skewb(JFrame frame, int dim,
			boolean orient, boolean practice,
			boolean mono, boolean reverse, String names,
			Color c0, Color c1, Color c2,
			Color c3, Color c4, Color c5,
			Color borderc, Color fg, Color bg, int delay) {
		this.frame = frame;
		this.dim = dim;
		this.orientChoice = orient;
		this.practiceChoice = practice;
		this.mono = mono;
		this.reverse = reverse;
		this.faceNames = names;
		this.faceColor[0] = c0;
		this.faceColor[1] = c1;
		this.faceColor[2] = c2;
		this.faceColor[3] = c3;
		this.faceColor[4] = c4;
		this.faceColor[5] = c5;
		this.currentForeground = this.borderColor = borderc;
		this.foreground = fg; this.background = bg;
		this.delay = delay;
		skewb2D = new Skewb2D(frame, this);
		skewb3D = new Skewb3D(frame, this);
		undo = ((SkewbFrame) frame).undo;
		redo = ((SkewbFrame) frame).redo;
		if (dim == 2)
			skewb2D.firstPaint = true;
		else if (dim == 3)
			skewb3D.firstPaint = true;
		this.addMouseListener(this);
		this.addMouseMotionListener(this);
		this.addMouseWheelListener(this);
		this.addKeyListener(this);
		setCursor(new Cursor(Cursor.HAND_CURSOR));
	}

	@Override
	protected String paramString() {
		return super.paramString() +
		",orient=" + orientChoice +
		",practice=" + practiceChoice +
		",mono=" + mono +
		",reverse=" + reverse +
		",names=" + faceNames +
		",color0=" + faceColor[0] +
		",color1=" + faceColor[1] +
		",color2=" + faceColor[2] +
		",color3=" + faceColor[3] +
		",color4=" + faceColor[4] +
		",color5=" + faceColor[5] +
		",color6=" + faceColor[6] +
		",color7=" + faceColor[7] +
		",borderColor=" + borderColor +
		",foreground=" + foreground +
		",background=" + background +
		",delay=" + delay;
	}

	void checkPieces() {
		if (delay < 0) {
			System.out.println(delay +
			" delay cannot be negative, taking absolute value");
			delay = -delay;
		}
	}

	boolean checkStarted() {
		return started;
	}

	boolean checkSolved() {
		for (int face = 0; face < MAX_FACES; face++) {
			int tempFace = cubeLoc[face][0].face;
			int tempRotation = cubeLoc[face][0].rotation;

			for (int position = 1; position < MAX_CUBES;
					position++) {
				if (tempFace != /* face */
						cubeLoc[face][position].face ||
						(orientChoice && tempRotation != /* STRT * MAX_ORIENT */
						cubeLoc[face][position].rotation))
					return false;
			}
		}
		return true;
	}

	void drawDiamond(int face, int offsetPosition) {
		if (dim == 2)
			skewb2D.drawDiamond(face, offsetPosition);
		else if (dim == 3)
			skewb3D.drawDiamond(face, offsetPosition);
	}

	void drawTriangle(int face, int position, int offsetPosition) {
		if (dim == 2)
			skewb2D.drawTriangle(face, position, offsetPosition);
		else if (dim == 3)
			skewb3D.drawTriangle(face, position, offsetPosition);
	}

	void drawFace(int face) {
		drawDiamond(face, 0);
		for (int position = 0; position < MAX_ORIENT; position++)
			drawTriangle(face, position, 0);
	}

	void drawAllPieces() {
		for (int face = 0; face < MAX_FACES; face++)
			drawFace(face);
		solve.init();
	}

	void moveNoPieces() {
		((SkewbFrame) frame).callback(
			SkewbInterface.ACTION_ILLEGAL, dim);
	}

	void rotateFace(int face, int direction) {
		/* Read Face */
		for (int corner = 0; corner < MAX_ORIENT; corner++) {
			faceLoc[corner].face = cubeLoc[face][corner].face;
			faceLoc[corner].rotation = cubeLoc[face][corner].rotation;
		}
		/* Write Face */
		for (int corner = 0; corner < MAX_ORIENT; corner++) {
			cubeLoc[face][corner].face = (direction == CW) ?
				faceLoc[(corner + MAX_ORIENT - 1) % MAX_ORIENT].face :
				faceLoc[(corner + 1) % MAX_ORIENT].face;
			cubeLoc[face][corner].rotation =
				(cubeLoc[face][corner].rotation + direction) %
				MAX_ORIENT;
			drawTriangle(face, corner, 0);
		}
		cubeLoc[face][MAX_ORIENT].rotation =
			(cubeLoc[face][MAX_ORIENT].rotation + direction) %
			MAX_ORIENT;
		drawDiamond(face, 0);
	}

	void readFace(int face, int h) {
		System.arraycopy(cubeLoc[face], 0, rowLoc[h], 0, MAX_CUBES);
	}

	void writeFace(int face, int rotate, int h) {
		int newCorner;

		for (int corner = 0; corner < MAX_ORIENT; corner++) {
			newCorner = (corner + rotate) % MAX_ORIENT;
			cubeLoc[face][newCorner] = rowLoc[h][corner];
			cubeLoc[face][newCorner].rotation =
				(cubeLoc[face][newCorner].rotation + rotate) %
				MAX_ORIENT;
			drawTriangle(face, (corner + rotate) % MAX_ORIENT, 0);
		}
		cubeLoc[face][MAX_ORIENT] = rowLoc[h][MAX_ORIENT];
		cubeLoc[face][MAX_ORIENT].rotation =
			(cubeLoc[face][MAX_ORIENT].rotation + rotate) %
			MAX_ORIENT;
		drawDiamond(face, 0);
	}

	void readDiagonal(int face, int corner, int orient, int size) {
		if (size == MINOR)
			minorLoc[orient] = cubeLoc[face][corner];
		else {			/* size == MAJOR */
			for (int g = 1; g < MAX_ORIENT; g++)
				majorLoc[orient][g - 1] =
					cubeLoc[face][(corner + g) %
					MAX_ORIENT];
			majorLoc[orient][MAX_ORIENT - 1] =
				cubeLoc[face][MAX_ORIENT];
		}
	}

	void rotateDiagonal(int rotate, int orient, int size) {
		if (size == MINOR)
			minorLoc[orient].rotation =
				(minorLoc[orient].rotation + rotate) %
				MAX_ORIENT;
		else			/* size == MAJOR */
			for (int g = 0; g < MAX_ORIENT; g++)
				majorLoc[orient][g].rotation =
					(majorLoc[orient][g].rotation +
					rotate) % MAX_ORIENT;
	}

	void writeDiagonal(int face, int corner, int orient, int size) {
		if (size == MINOR) {
			cubeLoc[face][corner] = minorLoc[orient];
			drawTriangle(face, corner, 0);
		} else {		/* size == MAJOR */
			cubeLoc[face][MAX_ORIENT] =
				majorLoc[orient][MAX_ORIENT - 1];
			drawDiamond(face, 0);
			for (int g = 1; g < MAX_ORIENT; g++) {

				int h = (corner + g) % MAX_ORIENT;
				cubeLoc[face][h] = majorLoc[orient][g - 1];
				drawTriangle(face, h, 0);
			}
		}
	}

	void moveFacesOrth(int face, int direction) {
		int newFace, newDirection, k, rotate;
		int oldFace = face, dir = direction;

		rotateFace(faceToRotate[oldFace][dir % MAX_ORIENT],
			CW);
		rotateFace(faceToRotate[oldFace][(dir + 2) % MAX_ORIENT],
			CCW);
		readFace(oldFace, 0);
		for (k = 1; k <= MAX_ORIENT; k++) {
			newFace = slideNextFace[oldFace][dir %
				MAX_ORIENT].face;
			rotate = slideNextFace[oldFace][dir %
				MAX_ORIENT].rotation; 
			newDirection = (rotate + dir) % MAX_ORIENT;
			if (k != MAX_ORIENT)
				readFace(newFace, k);
			writeFace(newFace, rotate, k - 1);
			oldFace = newFace;
			dir = newDirection;
		}
	}

	void movePieces(int face, int position, int direction) {
		int oldFace = face, pos = position, dir = direction;

		if (dir < 2 * MAX_ORIENT) {
			/* position as MAX_ORIENT is ambiguous */
			for (int size = MINOR; size <= MAJOR; size++) {
				readDiagonal(oldFace, pos, 0, size);
				for (int k = 1; k <= MAX_ROTATE; k++) {
					currentFace = slideNextRow[oldFace][pos][dir / 2].face;
					currentRotate = slideNextRow[oldFace][pos][dir / 2].rotation %
						MAX_ORIENT;
					currentDirection = (currentRotate + dir) % MAX_ORIENT;
					currentPosition = (currentRotate + pos) % MAX_ORIENT;
					if (k != MAX_ROTATE)
						readDiagonal(currentFace, currentPosition, k, size);
					rotateDiagonal(currentRotate, k - 1, size);
					writeDiagonal(currentFace, currentPosition, k - 1, size);
					oldFace = currentFace;
					pos = currentPosition;
					dir = currentDirection;
				}
				if (size == MINOR) {
					currentFace = minToMaj[oldFace][pos].face;
					currentRotate = minToMaj[oldFace][pos].rotation % MAX_ORIENT;
					dir = (currentRotate + dir) % MAX_ORIENT;
					pos = (pos + currentRotate + 2) % MAX_ORIENT;
					oldFace = currentFace;
				}
			}
		} else {
			moveFacesOrth(oldFace, dir);
		}
	}

	void moveControlCb(int face, int position, int direction) {
		int oldFace = face, pos = position, dir = direction;
		int newFace, rotate;

		if (currentDirection >= MAX_ORIENT) {
			newFace = face;
			if (direction == CW || direction == CCW) {
				newFace = rotateToRow[face].face;
				dir = 2 * MAX_ORIENT + ((direction == CCW) ?
					(rotateToRow[face].rotation + 2) :
					rotateToRow[face].rotation) % MAX_ORIENT;
			}
			((SkewbFrame) frame).callback(
				SkewbInterface.ACTION_CONTROL, dim,
				newFace, pos, dir);
			moveFacesOrth(newFace, dir);
			return;
		}
		movePieces(oldFace, pos, dir);
		((SkewbFrame) frame).callback(
			SkewbInterface.ACTION_CONTROL, dim,
			oldFace, pos, dir);
		newFace = minToMaj[oldFace][pos].face;
		rotate = minToMaj[oldFace][pos].rotation % MAX_ORIENT;
		dir = (rotate + dir) % MAX_ORIENT;
		pos = (pos + rotate + 2) % MAX_ORIENT;
		movePieces(newFace, pos, dir);
		((SkewbFrame) frame).callback(
			SkewbInterface.ACTION_CONTROL, dim,
			newFace, pos, dir);
	}

	void moveShiftCb(int face, int position, int direction) {
		int oldFace = face, pos = position, dir = direction;
		int newFace, rotate;

		if (currentDirection >= 2 * MAX_ORIENT) {
			SkewbLocPos newPos = orthToDiag[oldFace][pos]
				[dir % MAX_ORIENT];
			oldFace = newPos.face;
			pos = newPos.position;
			dir = newPos.direction;
		}
		movePieces(oldFace, pos, dir);
		((SkewbFrame) frame).callback(
			SkewbInterface.ACTION_CONTROL, dim,
			oldFace, pos, dir);
		newFace = minToMaj[oldFace][pos].face;
		rotate = minToMaj[oldFace][pos].rotation % MAX_ORIENT;
		dir = (rotate + dir) % MAX_ORIENT;
		pos = (pos + rotate + 2) % MAX_ORIENT;
		movePieces(newFace, pos, dir);
		((SkewbFrame) frame).callback(
			SkewbInterface.ACTION_CONTROL, dim,
			newFace, pos, dir);
	}

	void moveAltCb(int face, int position, int direction) {
		movePieces(face, position, direction);
		((SkewbFrame) frame).callback(
			SkewbInterface.ACTION_CONTROL, dim,
			face, position, direction);
	}

	void movePuzzle(int face, int position, int direction, int control) {
		int oldFace = face, pos = position, dir = direction;

		switch (control) {
		case 3:
			moveShiftCb(oldFace, pos, dir);
			break;
		case 2:
			moveAltCb(oldFace, pos, dir);
			break;
		case 1:
			moveControlCb(oldFace, pos, dir);
			break;
		default:
			SkewbLocPos newpos;
			if (dir >= 2 * MAX_ORIENT) {
				newpos = orthToDiag[oldFace][pos][dir % MAX_ORIENT];
				oldFace = newpos.face;
				pos = newpos.position;
				dir = newpos.direction;
			}
			movePieces(oldFace, pos, dir);
			((SkewbFrame) frame).callback(
				SkewbInterface.ACTION_MOVED, dim,
				oldFace, pos, dir);
			break;
		}
		undo.setMove(dir, control, oldFace, pos);
		redo.flushMoves();
		if (((SkewbFrame) frame).getToggleSound())
			((SkewbFrame) frame).playMoveAudio();
	}

	void movePuzzleDelay(int face, int position, int direction,
			int control, int fast) {
		movePuzzle(face, position, direction, control);
		try {
			Thread.sleep(64 * delay / fast);
		} catch (InterruptedException e) {
			// ignore
		}
	}

	void selectPieces(/*int x, int y*/) {
		if (nextControl == 0 && !practiceChoice && checkSolved())
			moveNoPieces();
		else if (nextFace >= 0) {
			int count;

			if (nextFace == currentFace) {
				if (currentPosition == MAX_ORIENT)
					count = -1;
				else
					// next sets currentDirection
					count = checkMoveDir(currentPosition,
						nextPosition);
				/* count == 2 too annoying */
				if (count == 2)
					count = -1;
			} else {
				currentDirection = facesToDirection[currentFace][nextFace];
				currentPosition = facestoPosition[currentFace][nextFace][currentPosition];
				count = nextFace;
				nextFace = currentFace;
				currentFace = count;
				if (currentPosition >= 0)
					count = 1;
				else
					count = 0;
			}
			if (count == 1) {
				movePuzzle(currentFace, currentPosition,
					currentDirection, nextControl);
				started = true;
				if (nextControl == 0 && checkSolved()) {
					((SkewbFrame) frame).callback(
						SkewbInterface.ACTION_SOLVED,
						dim);
				}
			} else if (count == 0)
				moveNoPieces();
		}
	}

	int checkMoveDir(int position1, int position2) {
		if (position1 == position2)
			return 2;
		if ((position1 - position2 + MAX_ORIENT) % 2 == 0)
			return 0;
		switch (position1) {
		case 0:
			currentDirection = ((position2 == 1) ? 2 : 3);
			break;
		case 1:
			currentDirection = ((position2 == 2) ? 3 : 0);
			break;
		case 2:
			currentDirection = ((position2 == 3) ? 0 : 1);
			break;
		case 3:
			currentDirection = ((position2 == 0) ? 1 : 2);
			break;
		default:
			return 0;
		}
		currentDirection += 2 * MAX_ORIENT;
		return 1;
	}

	boolean narrowSelection() {
		// sets currentFace, currentPosition, currentDirection
		if (dim == 2)
			skewb2D.narrowSelection();
		else if (dim == 3)
			skewb3D.narrowSelection();
		return true;
	}

	boolean positionPieces(int x, int y, int direction) {
		// sets currentFace, currentPosition
		if (dim == 2) {
			currentFace = skewb2D.positionToPiece(x, y);
		} else if (dim == 3) {
			currentFace = skewb3D.positionToPiece(x, y);
		}
		if (currentFace < 0)
			return false;
		currentDirection = direction;
		narrowSelection(); // sets currentFace, currentPosition, currentDirection
		return true;
	}

	void movePuzzleInput(int x, int y, int direction,
			boolean shift, boolean control, boolean alt) {
		int ctrl;

		if (!practiceChoice && !control && !alt && checkSolved()) {
			moveNoPieces();
			return;
		}
		if (!positionPieces(x, y, direction)) // sets currentFace, currentPosition, currentDirection
			return;
		if (alt)
			ctrl = 2;
		else
			ctrl = ((control) ? 1 : 0);
		if (ctrl != 1 && currentPosition == MAX_ORIENT)
			return;
		movePuzzle(currentFace, currentPosition, currentDirection, ctrl);
		if (ctrl == 0 && checkSolved()) {
			((SkewbFrame) frame).callback(
				SkewbInterface.ACTION_SOLVED, dim);
		}
	}

	void resetPieces() {
		cubeLoc = new SkewbLoc[MAX_FACES][MAX_CUBES];
		startLoc = new SkewbLoc[MAX_FACES][MAX_CUBES];
		rowLoc = new SkewbLoc[MAX_ORIENT][MAX_CUBES];
		minorLoc = new SkewbLoc[MAX_ORIENT];
		majorLoc = new SkewbLoc[MAX_ORIENT][MAX_ORIENT];
		faceLoc = new SkewbLoc[MAX_CUBES];
		for (int face = 0; face < MAX_FACES; face++) {
			for (int position = 0; position < MAX_CUBES; position++) {
				startLoc[face][position] = new SkewbLoc(face,
					STRT - MAX_ORIENT);
				cubeLoc[face][position] = new SkewbLoc(face,
					STRT - MAX_ORIENT);
			}
		}
		for (int orient = 0; orient < MAX_ORIENT; orient++) {
			minorLoc[orient] = new SkewbLoc(0, 0);
			for (int position = 0; position < MAX_CUBES; position++)
				rowLoc[orient][position] = new SkewbLoc(0, 0);
			for (int dir = 0; dir < MAX_ORIENT; dir++)
				majorLoc[orient][dir] = new SkewbLoc(0, 0);
			faceLoc[orient] = new SkewbLoc(0, STRT - MAX_ORIENT);
		}
		faceLoc[MAX_ORIENT] = new SkewbLoc(0, STRT - MAX_ORIENT);
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		currentFace = -1;
		currentPosition = -1;
		currentDirection = -1;
		started = false;
		cheat = false;
		skewb2D.randomPaints = 0;
		skewb3D.randomPaints = 0;
	}

	void practicePieces() {
		((SkewbFrame) frame).callback(
			SkewbInterface.ACTION_PRACTICE, dim);
	}

	void randomizePieces() {
		if (practiceChoice)
			practicePieces();
		((SkewbFrame) frame).callback(
			SkewbInterface.ACTION_RESET, dim);
		int big = MAX_CUBES * 3;
		if (dim == 2)
			skewb2D.randomPaints = java.lang.Math.min(big,
				MAX_ITERATIONS) + generator.nextInt(2);
		else if (dim == 3)
			skewb3D.randomPaints = java.lang.Math.min(big,
				MAX_ITERATIONS) + generator.nextInt(2);
		repaint();
	}

	void randomizingPieces() {
		int position, randomDirection;
		int face = generator.nextInt(MAX_FACES);

		position = generator.nextInt(MAX_ORIENT);
		randomDirection = ((generator.nextInt(2) == 1) ? position + 1 :
			position + 3) % MAX_ORIENT;
		movePuzzle(face, position, randomDirection, 0);
		if (dim == 2)
			skewb2D.randomPaints--;
		else if (dim == 3)
			skewb3D.randomPaints--;
		try {
			Thread.sleep(delay);
		} catch (InterruptedException e) {
			// ignore
		}
		if (dim == 2 && skewb2D.randomPaints > 0) {
			repaint();
		} else if (dim == 3 && skewb3D.randomPaints > 0) {
			repaint();
		} else {
			undo.flushMoves();
			redo.flushMoves();
			getStartPosition();
			((SkewbFrame) frame).callback(
				SkewbInterface.ACTION_RANDOMIZE, dim);
			started = false;
			cheat = false;
			((SkewbFrame) frame).callback(
				SkewbInterface.ACTION_CHEAT, dim);
			currentFace = -1;
			if (checkSolved())
				((SkewbFrame) frame).callback(
					SkewbInterface.ACTION_SOLVED, dim);
		}
	}

	public int getDelay() {
		return delay;
	}

	public void setDelay(int delay) {
		this.delay = delay;
	}

	public boolean getCheat() {
		return cheat;
	}

	public void setCheat(boolean cheater) {
		cheat = cheater;
	}

	public void setStart(boolean start) {
		started = start;
	}

	boolean setSkewbOrient(boolean newValue) {
		if (orientChoice != newValue) {
			orientChoice = newValue;
			resetPieces();
			if (dim == 2)
				skewb2D.resizePaint = true;
			else if (dim == 3)
				skewb3D.resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setSkewbPractice(boolean newValue) {
		if (practiceChoice != newValue) {
			practiceChoice = newValue;
			resetPieces();
			if (dim == 2)
				skewb2D.resizePaint = true;
			else if (dim == 3)
				skewb3D.resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	public void speedUpPuzzle() {
		delay -= 10;
		if (delay < 0)
			delay = 0;
		((SkewbFrame) frame).callback(SkewbInterface.ACTION_SPEED, dim);
	}

	public void slowDownPuzzle() {
		delay += 10;
		((SkewbFrame) frame).callback(SkewbInterface.ACTION_SPEED, dim);
	}

	void showMessage(String msg) {
		((SkewbFrame) frame).showMessage(msg);
	}

	public void initializePuzzle() {
		checkPieces();
		solve = new SkewbSolve(frame, this);
		resetPieces();
		generator = new Random(System.nanoTime());
		rotateToRow[0] = new SkewbLoc(1, LEFT);
		rotateToRow[1] = new SkewbLoc(0, BOTTOM);
		rotateToRow[2] = new SkewbLoc(0, RIGHT);
		rotateToRow[3] = new SkewbLoc(0, TOP);
		rotateToRow[4] = new SkewbLoc(1, RIGHT);
		rotateToRow[5] = new SkewbLoc(1, LEFT);
		slideNextRow[0][0][0] = new SkewbLoc(2, CW);
		slideNextRow[0][0][1] = new SkewbLoc(1, HALF);
		slideNextRow[0][1][0] = new SkewbLoc(5, CCW);
		slideNextRow[0][1][1] = new SkewbLoc(1, STRT);
		slideNextRow[0][2][0] = new SkewbLoc(3, STRT);
		slideNextRow[0][2][1] = new SkewbLoc(5, CW);
		slideNextRow[0][3][0] = new SkewbLoc(3, HALF);
		slideNextRow[0][3][1] = new SkewbLoc(2, CCW);
		slideNextRow[1][0][0] = new SkewbLoc(4, STRT);
		slideNextRow[1][0][1] = new SkewbLoc(5, CW);
		slideNextRow[1][1][0] = new SkewbLoc(0, STRT);
		slideNextRow[1][1][1] = new SkewbLoc(5, CCW);
		slideNextRow[1][2][0] = new SkewbLoc(2, CCW);
		slideNextRow[1][2][1] = new SkewbLoc(0, HALF);
		slideNextRow[1][3][0] = new SkewbLoc(2, CW);
		slideNextRow[1][3][1] = new SkewbLoc(4, HALF);
		slideNextRow[2][0][0] = new SkewbLoc(4, CW);
		slideNextRow[2][0][1] = new SkewbLoc(1, CCW);
		slideNextRow[2][1][0] = new SkewbLoc(0, CCW);
		slideNextRow[2][1][1] = new SkewbLoc(1, CW);
		slideNextRow[2][2][0] = new SkewbLoc(3, CCW);
		slideNextRow[2][2][1] = new SkewbLoc(0, CW);
		slideNextRow[2][3][0] = new SkewbLoc(3, CW);
		slideNextRow[2][3][1] = new SkewbLoc(4, CCW);
		slideNextRow[3][0][0] = new SkewbLoc(4, HALF);
		slideNextRow[3][0][1] = new SkewbLoc(2, CCW);
		slideNextRow[3][1][0] = new SkewbLoc(0, HALF);
		slideNextRow[3][1][1] = new SkewbLoc(2, CW);
		slideNextRow[3][2][0] = new SkewbLoc(5, CW);
		slideNextRow[3][2][1] = new SkewbLoc(0, STRT);
		slideNextRow[3][3][0] = new SkewbLoc(5, CCW);
		slideNextRow[3][3][1] = new SkewbLoc(4, STRT);
		slideNextRow[4][0][0] = new SkewbLoc(5, CW);
		slideNextRow[4][0][1] = new SkewbLoc(1, STRT);
		slideNextRow[4][1][0] = new SkewbLoc(2, CCW);
		slideNextRow[4][1][1] = new SkewbLoc(1, HALF);
		slideNextRow[4][2][0] = new SkewbLoc(3, HALF);
		slideNextRow[4][2][1] = new SkewbLoc(2, CW);
		slideNextRow[4][3][0] = new SkewbLoc(3, STRT);
		slideNextRow[4][3][1] = new SkewbLoc(5, CCW);
		slideNextRow[5][0][0] = new SkewbLoc(0, CW);
		slideNextRow[5][0][1] = new SkewbLoc(1, CW);
		slideNextRow[5][1][0] = new SkewbLoc(4, CCW);
		slideNextRow[5][1][1] = new SkewbLoc(1, CCW);
		slideNextRow[5][2][0] = new SkewbLoc(3, CW);
		slideNextRow[5][2][1] = new SkewbLoc(4, CW);
		slideNextRow[5][3][0] = new SkewbLoc(3, CCW);
		slideNextRow[5][3][1] = new SkewbLoc(0, CCW);
		minToMaj[0][0] = new SkewbLoc(3, CW);
		minToMaj[0][1] = new SkewbLoc(2, STRT);
		minToMaj[0][2] = new SkewbLoc(1, CCW);
		minToMaj[0][3] = new SkewbLoc(5, STRT);
		minToMaj[1][0] = new SkewbLoc(2, STRT);
		minToMaj[1][1] = new SkewbLoc(4, CCW);
		minToMaj[1][2] = new SkewbLoc(5, HALF);
		minToMaj[1][3] = new SkewbLoc(0, CW);
		minToMaj[2][0] = new SkewbLoc(3, STRT);
		minToMaj[2][1] = new SkewbLoc(4, STRT);
		minToMaj[2][2] = new SkewbLoc(1, STRT);
		minToMaj[2][3] = new SkewbLoc(0, STRT);
		minToMaj[3][0] = new SkewbLoc(5, HALF);
		minToMaj[3][1] = new SkewbLoc(4, CW);
		minToMaj[3][2] = new SkewbLoc(2, STRT);
		minToMaj[3][3] = new SkewbLoc(0, CCW);
		minToMaj[4][0] = new SkewbLoc(3, CCW);
		minToMaj[4][1] = new SkewbLoc(5, STRT);
		minToMaj[4][2] = new SkewbLoc(1, CW);
		minToMaj[4][3] = new SkewbLoc(2, STRT);
		minToMaj[5][0] = new SkewbLoc(3, HALF);
		minToMaj[5][1] = new SkewbLoc(0, STRT);
		minToMaj[5][2] = new SkewbLoc(1, HALF);
		minToMaj[5][3] = new SkewbLoc(4, STRT);
		slideNextFace[0][0] = new SkewbLoc(5, STRT);
		slideNextFace[0][1] = new SkewbLoc(3, CW);
		slideNextFace[0][2] = new SkewbLoc(2, STRT);
		slideNextFace[0][3] = new SkewbLoc(1, CCW);
		slideNextFace[1][0] = new SkewbLoc(0, CW);
		slideNextFace[1][1] = new SkewbLoc(2, STRT);
		slideNextFace[1][2] = new SkewbLoc(4, CCW);
		slideNextFace[1][3] = new SkewbLoc(5, HALF);
		slideNextFace[2][0] = new SkewbLoc(0, STRT);
		slideNextFace[2][1] = new SkewbLoc(3, STRT);
		slideNextFace[2][2] = new SkewbLoc(4, STRT);
		slideNextFace[2][3] = new SkewbLoc(1, STRT);
		slideNextFace[3][0] = new SkewbLoc(0, CCW);
		slideNextFace[3][1] = new SkewbLoc(5, HALF);
		slideNextFace[3][2] = new SkewbLoc(4, CW);
		slideNextFace[3][3] = new SkewbLoc(2, STRT);
		slideNextFace[4][0] = new SkewbLoc(2, STRT);
		slideNextFace[4][1] = new SkewbLoc(3, CCW);
		slideNextFace[4][2] = new SkewbLoc(5, STRT);
		slideNextFace[4][3] = new SkewbLoc(1, CW);
		slideNextFace[5][0] = new SkewbLoc(4, STRT);
		slideNextFace[5][1] = new SkewbLoc(3, HALF);
		slideNextFace[5][2] = new SkewbLoc(0, STRT);
		slideNextFace[5][3] = new SkewbLoc(1, HALF);
		orthToDiag[0][0][0] = new SkewbLocPos(3, 0, 1);
		orthToDiag[0][0][1] = new SkewbLocPos(5, 1, 0);
		orthToDiag[0][0][2] = new SkewbLocPos(3, 0, 3);
		orthToDiag[0][0][3] = new SkewbLocPos(5, 1, 2);
		orthToDiag[0][1][0] = new SkewbLocPos(3, 3, 0);
		orthToDiag[0][1][1] = new SkewbLocPos(2, 0, 1);
		orthToDiag[0][1][2] = new SkewbLocPos(3, 3, 2);
		orthToDiag[0][1][3] = new SkewbLocPos(2, 0, 3);
		orthToDiag[0][2][0] = new SkewbLocPos(1, 0, 3);
		orthToDiag[0][2][1] = new SkewbLocPos(2, 3, 0);
		orthToDiag[0][2][2] = new SkewbLocPos(1, 0, 1);
		orthToDiag[0][2][3] = new SkewbLocPos(2, 3, 2);
		orthToDiag[0][3][0] = new SkewbLocPos(1, 3, 2);
		orthToDiag[0][3][1] = new SkewbLocPos(5, 2, 1);
		orthToDiag[0][3][2] = new SkewbLocPos(1, 3, 0);
		orthToDiag[0][3][3] = new SkewbLocPos(5, 2, 3);
		orthToDiag[1][0][0] = new SkewbLocPos(2, 3, 0);
		orthToDiag[1][0][1] = new SkewbLocPos(0, 2, 1);
		orthToDiag[1][0][2] = new SkewbLocPos(2, 3, 2);
		orthToDiag[1][0][3] = new SkewbLocPos(0, 2, 3);
		orthToDiag[1][1][0] = new SkewbLocPos(2, 2, 3);
		orthToDiag[1][1][1] = new SkewbLocPos(4, 3, 0);
		orthToDiag[1][1][2] = new SkewbLocPos(2, 2, 1);
		orthToDiag[1][1][3] = new SkewbLocPos(4, 3, 2);
		orthToDiag[1][2][0] = new SkewbLocPos(5, 3, 2);
		orthToDiag[1][2][1] = new SkewbLocPos(4, 2, 3);
		orthToDiag[1][2][2] = new SkewbLocPos(5, 3, 0);
		orthToDiag[1][2][3] = new SkewbLocPos(4, 2, 1);
		orthToDiag[1][3][0] = new SkewbLocPos(5, 2, 1);
		orthToDiag[1][3][1] = new SkewbLocPos(0, 3, 2);
		orthToDiag[1][3][2] = new SkewbLocPos(5, 2, 3);
		orthToDiag[1][3][3] = new SkewbLocPos(0, 3, 0);
		orthToDiag[2][0][0] = new SkewbLocPos(3, 3, 0);
		orthToDiag[2][0][1] = new SkewbLocPos(0, 1, 0);
		orthToDiag[2][0][2] = new SkewbLocPos(3, 3, 2);
		orthToDiag[2][0][3] = new SkewbLocPos(0, 1, 2);
		orthToDiag[2][1][0] = new SkewbLocPos(3, 2, 3);
		orthToDiag[2][1][1] = new SkewbLocPos(4, 0, 1);
		orthToDiag[2][1][2] = new SkewbLocPos(3, 2, 1);
		orthToDiag[2][1][3] = new SkewbLocPos(4, 0, 3);
		orthToDiag[2][2][0] = new SkewbLocPos(1, 1, 0);
		orthToDiag[2][2][1] = new SkewbLocPos(4, 3, 0);
		orthToDiag[2][2][2] = new SkewbLocPos(1, 1, 2);
		orthToDiag[2][2][3] = new SkewbLocPos(4, 3, 2);
		orthToDiag[2][3][0] = new SkewbLocPos(1, 0, 3);
		orthToDiag[2][3][1] = new SkewbLocPos(0, 2, 1);
		orthToDiag[2][3][2] = new SkewbLocPos(1, 0, 1);
		orthToDiag[2][3][3] = new SkewbLocPos(0, 2, 3);
		orthToDiag[3][0][0] = new SkewbLocPos(5, 1, 2);
		orthToDiag[3][0][1] = new SkewbLocPos(0, 0, 3);
		orthToDiag[3][0][2] = new SkewbLocPos(5, 1, 0);
		orthToDiag[3][0][3] = new SkewbLocPos(0, 0, 1);
		orthToDiag[3][1][0] = new SkewbLocPos(5, 0, 1);
		orthToDiag[3][1][1] = new SkewbLocPos(4, 1, 2);
		orthToDiag[3][1][2] = new SkewbLocPos(5, 0, 3);
		orthToDiag[3][1][3] = new SkewbLocPos(4, 1, 0);
		orthToDiag[3][2][0] = new SkewbLocPos(2, 1, 0);
		orthToDiag[3][2][1] = new SkewbLocPos(4, 0, 1);
		orthToDiag[3][2][2] = new SkewbLocPos(2, 1, 2);
		orthToDiag[3][2][3] = new SkewbLocPos(4, 0, 3);
		orthToDiag[3][3][0] = new SkewbLocPos(2, 0, 3);
		orthToDiag[3][3][1] = new SkewbLocPos(0, 1, 0);
		orthToDiag[3][3][2] = new SkewbLocPos(2, 0, 1);
		orthToDiag[3][3][3] = new SkewbLocPos(0, 1, 2);
		orthToDiag[4][0][0] = new SkewbLocPos(3, 2, 3);
		orthToDiag[4][0][1] = new SkewbLocPos(2, 1, 0);
		orthToDiag[4][0][2] = new SkewbLocPos(3, 2, 1);
		orthToDiag[4][0][3] = new SkewbLocPos(2, 1, 2);
		orthToDiag[4][1][0] = new SkewbLocPos(3, 1, 2);
		orthToDiag[4][1][1] = new SkewbLocPos(5, 0, 1);
		orthToDiag[4][1][2] = new SkewbLocPos(3, 1, 0);
		orthToDiag[4][1][3] = new SkewbLocPos(5, 0, 3);
		orthToDiag[4][2][0] = new SkewbLocPos(1, 2, 1);
		orthToDiag[4][2][1] = new SkewbLocPos(5, 3, 0);
		orthToDiag[4][2][2] = new SkewbLocPos(1, 2, 3);
		orthToDiag[4][2][3] = new SkewbLocPos(5, 3, 2);
		orthToDiag[4][3][0] = new SkewbLocPos(1, 1, 0);
		orthToDiag[4][3][1] = new SkewbLocPos(2, 2, 1);
		orthToDiag[4][3][2] = new SkewbLocPos(1, 1, 2);
		orthToDiag[4][3][3] = new SkewbLocPos(2, 2, 3);
		orthToDiag[5][0][0] = new SkewbLocPos(3, 1, 2);
		orthToDiag[5][0][1] = new SkewbLocPos(4, 1, 0);
		orthToDiag[5][0][2] = new SkewbLocPos(3, 1, 0);
		orthToDiag[5][0][3] = new SkewbLocPos(4, 1, 2);
		orthToDiag[5][1][0] = new SkewbLocPos(3, 0, 1);
		orthToDiag[5][1][1] = new SkewbLocPos(0, 0, 1);
		orthToDiag[5][1][2] = new SkewbLocPos(3, 0, 3);
		orthToDiag[5][1][3] = new SkewbLocPos(0, 0, 3);
		orthToDiag[5][2][0] = new SkewbLocPos(1, 3, 2);
		orthToDiag[5][2][1] = new SkewbLocPos(0, 3, 0);
		orthToDiag[5][2][2] = new SkewbLocPos(1, 3, 0);
		orthToDiag[5][2][3] = new SkewbLocPos(0, 3, 2);
		orthToDiag[5][3][0] = new SkewbLocPos(1, 2, 1);
		orthToDiag[5][3][1] = new SkewbLocPos(4, 2, 1);
		orthToDiag[5][3][2] = new SkewbLocPos(1, 2, 3);
		orthToDiag[5][3][3] = new SkewbLocPos(4, 2, 3);
	}

	void exposePuzzle() {
		if (dim == 2)
			skewb2D.exposePuzzle();
		else if (dim == 3)
			skewb3D.exposePuzzle();
	}

	void selectPuzzle(int control, int x, int y) {
		if (dim == 2)
			skewb2D.selectPuzzle(control, x, y);
		else if (dim == 3)
			skewb3D.selectPuzzle(control, x, y);
	}

	void releasePuzzle(int control, int x, int y) {
		if (dim == 2)
			skewb2D.releasePuzzle(control, x, y);
		else if (dim == 3)
			skewb3D.releasePuzzle(control, x, y);
	}

	void randomizePuzzle() {
		if (!mouseDown)
			randomizePieces();
	}

	void randomizePuzzleWithQuery() {
		if (!started)
			randomizePuzzle();
	}

	public void getPuzzle(boolean incMoves) {
		this.stringSave = ((SkewbFrame) frame).readTextArea();
		if (this.stringSave == null || this.stringSave.equals("")) {
			if (incMoves)
				showMessage("Empty file to read");
			return;
		}
		getPuzzleString("buffer", incMoves);
	}

	public void getPuzzleFile(boolean incMoves) {
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		try (BufferedReader br = new BufferedReader(
				new FileReader(DATAFILE))) {
			StringBuilder puzzleData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null)
				puzzleData.append(lineData).append(NL);
			this.stringSave = puzzleData.toString();
			if (this.stringSave == null || this.stringSave.length() == 0) {
				if (incMoves)
					showMessage("Empty file " + DATAFILE);
				return;
			}
			getPuzzleString(DATAFILE, incMoves);
		} catch (Exception e) {
			//e.printStackTrace();
			if (incMoves)
				showMessage("Cannot read " + DATAFILE);
		}
	}

	public void getPuzzleString(String source, boolean incMoves) {
		try {
			int moves;
			boolean orient, practice;
			String symbol = SYMBOL;
			String newline = NL;
			int line = 0;

			token = this.stringSave.split(newline);
			String lineData = token[line];
			if (lineData == null || lineData.length() == 0) {
				if (incMoves)
					showMessage(source +
						" nothing in buffer");
				return;
			}
			orient = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.orientChoice != orient)
				((SkewbFrame) frame).callback(
					SkewbInterface.ACTION_ORIENTIZE, dim);
			lineData = token[++line];
			practice = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.practiceChoice != practice)
				((SkewbFrame) frame).callback(
					SkewbInterface.ACTION_PRACTICE, dim);
			lineData = token[++line];
			moves = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			line = scanStartPosition(++line, incMoves);
			if (line < 0)
				return;
			((SkewbFrame) frame).callback(
				SkewbInterface.ACTION_RESTORE, dim);
			setStartPosition();
			if (dim == 2)
				skewb2D.allPiecesPaint = true;
			else if (dim == 3)
				skewb3D.allPiecesPaint = true;
			repaint();
			if (!scanMoves(++line, moves, incMoves))
				return;
			if (incMoves)
				System.out.println(source + ": orient " + orient +
					", practice " + practice + ", moves " + moves);
			cheat = true;
			((SkewbFrame) frame).callback(
				SkewbInterface.ACTION_CHEAT, dim);
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			if (incMoves)
				showMessage("Corrupt input when reading");
		} catch (Exception e) {
			//e.printStackTrace();
			if (incMoves)
				showMessage("Bad input file");
		}
	}

	public void printPuzzle(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append("orient").append(symbol).append(space)
						.append((orientChoice) ? 1 : 0).append(newline);
		sb.append("practice").append(symbol).append(space)
						.append((practiceChoice) ? 1 : 0).append(newline);
		sb.append("moves").append(symbol).append(space)
						.append(undo.numMoves()).append(newline);
		printStartPosition(sb);
		printMoves(sb);
	}

	public void writePuzzle() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		((SkewbFrame) frame).writeTextArea(sb.toString());
	}

	public void writePuzzleFile() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				DATAFILE, false))) {
			pr.print(sb);
			System.out.println("Saved to " + DATAFILE);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Cannot write to " + DATAFILE);
		}
	}

	void undoPuzzle() {
		if (mouseDown || skewb2D.stackPaint || skewb3D.stackPaint ||
			skewb2D.randomPaints != 0 || skewb3D.randomPaints != 0 ||
			!undo.madeMoves())
			return;
		undo.getMove();
		currentDirection = undo.readMoveDirection();
		currentControl = undo.readMoveControl();
		currentFace = undo.readMoveFace();
		currentPosition = undo.readMovePosition();
		redo.setMove(currentDirection, currentControl,
			currentFace, currentPosition);
		// move in opposite direction
		if (currentDirection < 2 * MAX_ORIENT)
			currentDirection = (currentDirection < MAX_ORIENT) ?
				(currentDirection + MAX_ORIENT / 2) %
				MAX_ORIENT : 3 * MAX_ORIENT - currentDirection;
		else
			currentDirection = 5 * MAX_ORIENT - currentDirection;
		if (currentControl == 0) {
			/* if (checkmovePiecesDir(currentDirection)) */
			if (currentDirection >= 2 * MAX_ORIENT) {
				SkewbLocPos newPos = orthToDiag[currentFace]
					[currentPosition]
					[currentDirection % MAX_ORIENT];
				currentFace = newPos.face;
				currentPosition = newPos.position;
				currentDirection = newPos.direction;
			}
			if (dim == 2)
				skewb2D.stackPaint = true;
			else if (dim == 3)
				skewb3D.stackPaint = true;
			repaint();
			started = true;
			((SkewbFrame) frame).callback(
				SkewbInterface.ACTION_UNDO, dim,
				currentFace, currentPosition, currentDirection);
		} else {
			if (dim == 2)
				skewb2D.stackPaint = true;
			else if (dim == 3)
				skewb3D.stackPaint = true;
			repaint();
		} /* else currentDirection = currentFace = -1; */
	}

	void redoPuzzle() {
		if (mouseDown || skewb2D.stackPaint || skewb3D.stackPaint ||
			skewb2D.randomPaints != 0 || skewb3D.randomPaints != 0 ||
			!redo.madeMoves())
			return;
		redo.getMove();
		currentDirection = redo.readMoveDirection();
		currentControl = redo.readMoveControl();
		currentFace = redo.readMoveFace();
		currentPosition = redo.readMovePosition();
		undo.setMove(currentDirection, currentControl,
			currentFace, currentPosition);
		if (currentControl == 0) {
			/* if (checkmovePiecesDir(currentDirection)) */
			if (currentDirection >= 2 * MAX_ORIENT) {
				SkewbLocPos newPos = orthToDiag[currentFace]
					[currentPosition]
					[currentDirection % MAX_ORIENT];
				currentFace = newPos.face;
				currentPosition = newPos.position;
				currentDirection = newPos.direction;
			}
			if (dim == 2)
				skewb2D.stackPaint = true;
			else if (dim == 3)
				skewb3D.stackPaint = true;
			repaint();
			started = true;
			((SkewbFrame) frame).callback(
				SkewbInterface.ACTION_REDO, dim,
				currentFace, currentPosition, currentDirection);
		} else {
			if (dim == 2)
				skewb2D.stackPaint = true;
			else if (dim == 3)
				skewb3D.stackPaint = true;
			repaint();
		} /* else currentDirection = currentFace = -1; */
	}

	public void clearPuzzle() {
		if (mouseDown || skewb2D.randomPaints != 0 ||
				skewb3D.randomPaints != 0)
			return;
		resetPieces();
		drawAllPieces();
		((SkewbFrame) frame).callback(
			SkewbInterface.ACTION_CLEAR, dim);
		if (dim == 2)
			skewb2D.allPiecesPaint = true;
		else if (dim == 3)
			skewb3D.allPiecesPaint = true;
		repaint();
	}

	void solvePuzzle() {
		if (mouseDown || skewb2D.stackPaint || skewb3D.stackPaint ||
				skewb2D.randomPaints != 0 || skewb3D.randomPaints != 0)
			return;
		cheat = true;
		((SkewbFrame) frame).callback(
			SkewbInterface.ACTION_CHEAT, dim);
		solve.setFlag(true);
		try {
			Thread.sleep(200);
		} catch (InterruptedException e) {
			System.err.println("Interrupted");
		}
		//solve.solvePieces();
	}

	void orientizePuzzle() {
		if (mouseDown || skewb2D.randomPaints != 0 ||
				skewb3D.randomPaints != 0)
			return;
		((SkewbFrame) frame).callback(
			SkewbInterface.ACTION_ORIENTIZE, dim);
	}

	void practicePuzzle() {
		if (mouseDown || skewb2D.randomPaints != 0 ||
				skewb3D.randomPaints != 0)
			return;
		practicePieces();
	}

	void practicePuzzleWithQuery() {
		if (!started)
			practicePuzzle();
	}

	static void printPositions() {
		for (int face = 0; face < MAX_FACES; face++) {
			for (int position = 0; position < MAX_CUBES; position++) {
				int pos = startLoc[face][position].face;
				int rot = startLoc[face][position].rotation;

				if (pos < 100)
					System.out.print(" ");
				if (pos < 10 && pos >= 0)
					System.out.print(" ");
				System.out.print(pos + "," + rot);
				if (position == MAX_CUBES - 1) {
					System.out.println("");
				} else {
					System.out.print(" ");
				}
			}
			System.out.println("");
		}
	}

	void printCube() {
		for (int face = 0; face < MAX_FACES; face++) {
			for (int position = 0; position < MAX_CUBES; position++)
				System.out.print(
					cubeLoc[face][position].face + " " +
					cubeLoc[face][position].rotation + "  ");
			System.out.println("");
		}
		System.out.println("");
	}

	void printFace() {
		for (int position = 0; position < MAX_CUBES; position++)
			System.out.print(
				faceLoc[position].face + " " +
				faceLoc[position].rotation + "  ");
		System.out.println("");
	}


	void printRow(int orient, int size) {
		System.out.println("Row " + orient + ":");
		for (int g = 0; g < size; g++)
			System.out.print(
					rowLoc[orient][g].face + " " +
					rowLoc[orient][g].rotation +
					"  ");
		System.out.println("");
	}

	void getStartPosition() {
		for (int face = 0; face < MAX_FACES; face++)
			for (int position = 0; position < MAX_CUBES; position++) {
				startLoc[face][position].face =
					cubeLoc[face][position].face;
				startLoc[face][position].rotation =
					cubeLoc[face][position].rotation;
			}
	}

	void setStartPosition() {
		for (int face = 0; face < MAX_FACES; face++)
			for (int position = 0; position < MAX_CUBES; position++) {
				cubeLoc[face][position].face =
					startLoc[face][position].face;
				cubeLoc[face][position].rotation =
					startLoc[face][position].rotation;
			}
		// drawAll
	}

	int scanStartPosition(int location, boolean incMoves) {
		String lineData;
		int line = location;

		try {
			String space = "\\s+";
			int group = (orientChoice) ? 2 : 1;

			++line; //read startingPosition:
			for (int face = 0; face < MAX_FACES; face++) {
				line++;
				if (line >= token.length) {
					if (incMoves)
						System.out.println("Short data in StartPosition");
					return -1;
				}
				lineData = token[line];
				String[] st = lineData.split(space);
				int offset = ("".equals(st[0])) ? 1 : 0;
				for (int position = 0; position < MAX_CUBES; position++)
				{
					int num = Integer.parseInt(st[group * position + offset].trim());
					cubeLoc[face][position].face =
					startLoc[face][position].face = num;
					if (orientChoice) {
						num = Integer.parseInt(st[group * position + offset + 1].trim());
						cubeLoc[face][position].rotation =
						startLoc[face][position].rotation = num;
					}
				}
			}
			//blank line read
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			if (incMoves)
				showMessage("Corrupt input in StartPosition");
			return -1;
		}
		return line;
	}

	boolean scanMoves(int location, int moves, boolean incMoves) {
		Graphics g = getGraphics();

		if (g == null)
			return false;
		try {
			int l;
			int line = location;
			String lineData;
			String symbol = SYMBOL;
			String space = "\\s+";

			++line; //read moves direction
			paint(g);
			for (l = 0; l < moves && ++line < token.length; l++) {
				lineData =  token[line];
				String[] st = lineData.substring(
					lineData.indexOf(symbol) + 2).split(space);
				currentDirection = Integer.parseInt(st[0].trim());
				currentControl = Integer.parseInt(st[1].trim());
				currentFace = Integer.parseInt(st[2].trim());
				currentPosition = Integer.parseInt(st[3].trim());
				{
					if (dim == 2 && currentControl == 0)
						skewb2D.scanPaint = true;
					else if (dim == 3)
						skewb3D.scanPaint = true;
					undo.setMove(currentDirection,
						currentControl,
						currentFace, currentPosition);
					redo.flushMoves();
					paint(g);
					if (incMoves && currentControl == 0)
						((SkewbFrame) frame).callback(
						  SkewbInterface.ACTION_MOVED,
						  dim);
				} /* else {
					currentDirection = -1;
					System.out.println(l +
						": move in direction " +
						currentDirection +
						", cannot be made.");
					return false;
				} */

			}
			if (l < moves) {
				if (incMoves)
					showMessage("Corrupt input in scanMoves, only " + l + " moves");
				return false;
			}
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			if (incMoves)
				showMessage("Corrupt input in scanMoves");
			return false;
		} finally {
			g.dispose();
		}
		return true;
	}

	void printStartPosition(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append(newline);
		sb.append("startingPosition").append(symbol).append(newline);
		for (int face = 0; face < MAX_FACES; face++) {
			for (int position = 0; position < MAX_CUBES; position++)
			{
				int temp = startLoc[face][position].face;
				if (temp < 100)
					sb.append(space);
				if (temp < 10 && temp >= 0)
					sb.append(space);
				sb.append(space).append(temp);
				if (orientChoice) {
					temp = startLoc[face][position].rotation;
					sb.append(space);
					if (temp < 10 && temp >= 0)
						sb.append(space);
					sb.append(space).append(temp);
				}
			}
			sb.append(newline);
		}
		sb.append(newline);
	}

	void printMoves(StringBuilder sb) {
		int direction, control;
		int face, position;
		int nMoves = undo.numMoves();
		String symbol = SYMBOL;
		String space = "\t";
		String newline = NL;

		sb.append("moves").append(space)
						.append("dir").append(space)
						.append("control").append(space)
						.append("face").append(space)
						.append("position").append(newline);
		while (undo.madeMoves()) {
			undo.getMove();
			direction = undo.readMoveDirection();
			control = undo.readMoveControl();
			face = undo.readMoveFace();
			position = undo.readMovePosition();
			redo.setMove(direction, control, face, position);
		}
		// apply all moves back to the stack since
		// we emptied it above!!
		for (int i = 0; i < nMoves; i++) {
			redo.getMove();
			direction = redo.readMoveDirection();
			control = redo.readMoveControl();
			face = redo.readMoveFace();
			position = redo.readMovePosition();
			sb.append(i + 1).append(symbol).append(space)
								.append(direction).append(space)
								.append(control).append(space)
								.append(face).append(space)
								.append(position).append(newline);
			undo.setMove(direction, control, face, position);
		}
	}

	@Override
	public void paint(Graphics g) {
		if (dim == 2)
			skewb2D.paint(g);
		else if (dim == 3)
			skewb3D.paint(g);
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}

	public boolean processKey(char key) {
		switch (key) {
		case '>':
		case '.':
			speedUpPuzzle();
			return true;
		case '<':
		case ',':
			slowDownPuzzle();
			return true;
		case '@':
			((SkewbFrame) frame).toggleSound();
			return true;
		case '?':
			((SkewbFrame) frame).descriptionHelp();
			return true;
		case '!':
			((SkewbFrame) frame).featuresHelp();
			return true;
		case '^':
			((SkewbFrame) frame).referencesHelp();
			return true;
		case 'a':
		case 'A':
			((SkewbFrame) frame).aboutHelp();
			return true;
		case '*':
			((SkewbFrame) frame).showBuffer();
			return true;
		default:
		}
		if (solve != null && solve.getFlag()) {
			solve.setFlag(false);
			return false;
		}
		switch (key) {
		case 'g':
			((SkewbFrame) frame).getPuzzle();
			break;
		case 'w':
			writePuzzle();
			break;
		case 'G':
			((SkewbFrame) frame).getPuzzleFile();
			break;
		case 'W':
			writePuzzleFile();
			break;
		case 'u':
		case 'U':
			undoPuzzle();
			break;
		case 'r':
		case 'R':
			redoPuzzle();
			break;
		case 'c':
		case 'C':
			clearPuzzle();
			break;
		case 'z':
		case 'Z':
			randomizePuzzle();
			break;
		case 'p':
		case 'P':
			practicePuzzle();
			break;
		case 's':
		case 'S':
			solvePuzzle();
			break;
		case 'o':
		case 'O':
			orientizePuzzle();
			break;
		case 'q':
		case 'Q':
		case 'x':
		case 'X':
		case '\03': /* Control-C */
			((SkewbFrame) frame).quit();
			break;
		default:
			return false;
		}
		return true;
	}

	@Override
	public void mousePressed(MouseEvent event) {
		int modifiers = event.getModifiersEx();

		if (skewb2D.scanPaint || skewb3D.scanPaint ||
				skewb2D.stackPaint || skewb3D.stackPaint ||
				mouseDown)
			return;
		if (solve != null && solve.getFlag()) {
			solve.setFlag(false);
			return;
		}
		if (((modifiers & InputEvent.META_DOWN_MASK) == 0) &&
				((modifiers & InputEvent.ALT_DOWN_MASK) == 0)) {
			if (skewb2D.randomPaints == 0 &&
					skewb3D.randomPaints == 0) {
				selectPuzzle(((modifiers & InputEvent.ALT_DOWN_MASK) == 0) ?
					(((modifiers & InputEvent.CTRL_DOWN_MASK) == 0) ? 0 : 1) : 2,
					event.getX(), event.getY());
				mouseDown = true;
			}
		} else {
			if ((event.getClickCount() & 1) == 1)
				randomizePuzzleWithQuery();
			else
				randomizePuzzle();
		}
	}

	@Override
	public void mouseClicked(MouseEvent event) {
		/* Method required but mousePressed does work */
	}

	@Override
	public void mouseReleased(MouseEvent event) {
		int modifiers = event.getModifiersEx();

		if (mouseDown) { // Do not care which mouse button
			releasePuzzle(((modifiers & InputEvent.ALT_DOWN_MASK) == 0) ?
				(((modifiers & InputEvent.CTRL_DOWN_MASK) == 0) ? 0 : 1) : 2,
				event.getX(), event.getY()); mouseDown = false;
		}
	}

	@Override
	public void mouseEntered(MouseEvent event) {
		currentForeground = foreground;
		if (dim == 2) {
			/* skewb2D.requestFocus(); */
			skewb2D.framePaint = true;
		}
		else if (dim == 3) {
			/* skewb3D.requestFocus(); */
			skewb3D.framePaint = true;
		}
		mouseDown = false;
		repaint();
	}

	@Override
	public void mouseExited(MouseEvent event) {
		if (mouseDown && currentFace >= 0) {
			if (dim == 2)
				skewb2D.releasePaint = true;
			else if (dim == 3)
				skewb3D.releasePaint = true;
		}
		currentForeground = borderColor;
		if (dim == 2)
			skewb2D.framePaint = true;
		else if (dim == 3)
			skewb3D.framePaint = true;
		mouseDown = false;
		repaint();
	}

	@Override
	public void mouseDragged(MouseEvent event) {
		/* Method required but mousePressed does work */
	}

	@Override
	public void mouseMoved(MouseEvent event) {
		lastX = event.getX();
		lastY = event.getY();
	}

	@Override
	public void mouseWheelMoved(MouseWheelEvent event) {
		int notches = event.getWheelRotation();
		int modifiers = event.getModifiersEx();

		if (notches < 0) {
			movePuzzleInput(lastX, lastY, TOP,
				(modifiers & InputEvent.SHIFT_DOWN_MASK) != 0,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
		} else {
			movePuzzleInput(lastX, lastY, BOTTOM,
				(modifiers & InputEvent.SHIFT_DOWN_MASK) != 0,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
		}
	}

	@Override
	public void keyPressed(KeyEvent event) {
		if (event.getKeyCode() == KeyEvent.VK_ESCAPE)
			((SkewbFrame) frame).shuffleDown();
		else if (!processKey(event.getKeyChar())) {
			if (dim == 2)
				skewb2D.keyPressed(event);
			else if (dim == 3)
				skewb3D.keyPressed(event);
		}
	}

	@Override
	public void keyReleased(KeyEvent event) {
		/* Method required but keyPressed does work */
	}

	@Override
	public void keyTyped(KeyEvent event) {
		/* Method required but keyPressed does work */
	}
}

/**
 * The <code>SkewbLoc</code> class maintains the locations of
 * the pieces.
 */

class SkewbLoc {
	int		 face;
	int		 rotation;

	public SkewbLoc(int face, int rotation) {
		this.face = face;
		this.rotation = rotation;
	}
}

/**
 * The <code>SkewbLocPos</code> class maintains the locations and
 * direction of the pieces.
 */

class SkewbLocPos {
	int		 face;
	int		 position;
	int		 direction;

	public SkewbLocPos(int face, int position, int direction) {
		this.face = face;
		this.position = position;
		this.direction = direction;
	}
}
